from collections import namedtuple
import logging
import sys
import datetime as dt
from pathlib import Path
from unittest import mock

import pytest

sys.path.insert(0, Path(__file__).parent / 'webservice')
import webservice  # noqa: import not at top of file
from webservice.messages import MigrationError  # noqa: import not at top
from webservice.webservice import (  # noqa: import not at top of file
    check_files,
    merge,
    parse_config,
    run_action,
    wait_on_transfer,
    get_slurm_partition,
    get_slurm_nice,
    config,
)


@pytest.mark.requires_gpfs
def test_check_files():
    in_folder = '/gpfs/exfel/exp/CALLAB/202031/p900113/raw'
    runs = [9985, 9984]
    karabo_das = ['AGIPD06', 'AGIPD07']

    assert check_files(in_folder, runs, karabo_das)

    karabo_das = ['LPD06', 'LPD07']
    assert not check_files(in_folder, runs, karabo_das)

    runs = [1, 2]
    assert not check_files(in_folder, runs, karabo_das)

    with pytest.raises(PermissionError):
        in_folder = '/gpfs/maxwell/home/achilles'  # arbitrarily chosen
        check_files(in_folder, runs, karabo_das)


def test_merge():
    a = {'some': {'key': {'akey': 'avalue', 'number': 1}}}
    b = {'some': {'key': {'anotherkey': 'anothervalue', 'number': 5}},
         'completely': 'different'}

    ret = merge(a, b)
    expected = {'some': {'key': {'akey': 'avalue',
                                 'anotherkey': 'anothervalue',
                                 'number': 1}},
                'completely': 'different'}
    assert ret == expected


def test_parse_config():
    cmd = ['whatever']
    config = {'somebool': True,
              'notsomebool': False,
              'alist': [1, 2, 3],
              'some_empty_key': '""',
              'other_empty_key': "''",
              'brian': 'scone'}

    expected = ['whatever', '--somebool', '--alist', '1', '2', '3',
                '--some_empty_key', '', '--other_empty_key', '',
                '--brian', 'scone']

    config = parse_config(cmd, config)

    assert config == expected
    assert '--notsomebool' not in config

    with pytest.raises(ValueError):
        config = {'some key': 'value'}
        config = parse_config(cmd, config)

    with pytest.raises(ValueError):
        config = {'somekey': 'a value'}
        config = parse_config(cmd, config)


@pytest.mark.asyncio
@pytest.mark.parametrize("run_links_to", ["XFEL_GPFS_OFFLINE_RAW_CC", "DESY_DCACHE_RAW_CC"])
@pytest.mark.parametrize(
    "repositories",
    {
        (gpfs_online, gpfs_offline, dcache)
        for gpfs_online in [None, None, "XFEL_GPFS_ONLINE_RAW_CC"]
        for gpfs_offline in [None, None, "XFEL_GPFS_OFFLINE_RAW_CC"]
        for dcache in [None, None, "DESY_DCACHE_RAW_CC"]
    }
)
async def test_wait_on_transfer(
    tmp_path: Path,
    run_links_to: str,
    repositories: list,
):
    tmp_root = tmp_path
    inst = "SCS"
    cycle = 202202
    prop_no = 1111
    run_no = 1

    repositories = list(filter(None, repositories))
    expected_error = run_links_to not in repositories

    gpfs_prefix = tmp_root / "gpfs/exfel/d"
    gpfs_dir = gpfs_prefix / f"raw/{inst}/{cycle}/p{prop_no}"
    gpfs_dir.mkdir(parents=True)

    pnfs_prefix = tmp_root / "pnfs/xfel.eu/exfel/archive/XFEL"
    pnfs_dir = pnfs_prefix  / f"raw/{inst}/{cycle}/p{prop_no}"
    pnfs_dir.mkdir(parents=True)

    raw_path = tmp_root / f"gpfs/exfel/exp/{inst}/{cycle}/p{prop_no:06d}/raw"
    raw_path.parent.mkdir(parents=True)
    if run_links_to == "XFEL_GPFS_OFFLINE_RAW_CC":
        raw_path.symlink_to(gpfs_dir)
    else:
        raw_path.symlink_to(pnfs_dir)

    run_path = raw_path / "r{run_no:04d}"
    run_path.mkdir(parents=True)

    response = mock.Mock()
    response.status_code = 200
    response.json = lambda: {"runs": [{"repositories": {r: None for r in repositories}}]}
    client = mock.Mock()
    client.get_runs_by_proposal_number_api = mock.Mock(return_value=response)

    _run_path = mock.Mock(spec=run_path)
    _run_path.resolve = lambda: Path("/") / run_path.resolve().relative_to(tmp_path)

    try:
        await wait_on_transfer(
            client,
            run=run_no,
            proposal=str(prop_no),
            run_dir=_run_path,
            max_tries=1,
            sleep_completion=0,
        )
    except MigrationError as e:
        if not expected_error:
            raise e


@pytest.mark.parametrize(
    "repositories",
    {
        (gpfs_online, gpfs_offline, dcache)
        for gpfs_online in [None, None, "XFEL_GPFS_ONLINE_RAW_CC"]
        for gpfs_offline in [None, None, "XFEL_GPFS_OFFLINE_RAW_CC"]
        for dcache in [None, None, "DESY_DCACHE_RAW_CC"]
    }
)
@pytest.mark.asyncio
async def test_wait_on_transfer_invalid_run_dir(caplog, repositories):
    response = mock.Mock()
    response.status_code = 200
    response.json = lambda: {"runs": [{"repositories": {r: None for r in repositories}}]}
    client = mock.Mock()
    client.get_runs_by_proposal_number_api = mock.Mock(return_value=response)

    caplog.set_level(logging.WARNING)

    res = await wait_on_transfer(
        client,
        run=0,
        proposal=str(10),
        run_dir=Path("/invalid/run/dir"),
        max_tries=1,
        sleep_completion=0,
    )

    assert res is None
    assert "is not relative to" in caplog.text

@pytest.mark.asyncio
@pytest.mark.parametrize(
    "mock_json, status_code, exception_class, exception_match",
    [
        ([],
         200,
         ValueError,
         r"FAILED:.*empty response.*"),
        ([],
         404,
         ValueError,
         r"FAILED:.*MyMDC.*404.*"),
        ({'runs': [{'repositories': {}}]},
         200,
         MigrationError,
         r"FAILED:.*migration issue.*"),
    ],
)
async def test_wait_on_transfer_exceptions(
    mock_json, status_code, exception_class, exception_match
):
    response = mock.Mock()
    response.status_code = status_code
    response.json = mock.Mock(return_value=mock_json)
    client = mock.Mock()
    client.get_runs_by_proposal_number_api = mock.Mock(
        return_value=response)

    with pytest.raises(exception_class, match=exception_match):
        await wait_on_transfer(
            client,
            run=1,
            proposal="1",
            max_tries=1,
            sleep_completion=0,
            run_dir=Path("/gpfs/exfel/d"),  # valid path required to not exit early
        )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'mode, cmd, retcode, expected',
    [
        ('prod', ['CORRECT', '1', '2', '3', '4'], 1, "failed: failed executing"),  # noqa
        ('prod', ['DARK', '1', '2', '3', '4'], 1, "failed: failed executing"),
        ('prod', ['CORRECT', '1', '2', '3', '4'], 0, "success: started correction"),  # noqa
        ('prod', ['DARK', '1', '2', '3', '4'], 0, "success: started dark characterization"),  # noqa
        ('sim', ['CORRECT', '1', '2', '3', '4'], 1, "success: simulated"),
        ('sim', ['DARK', '1', '2', '3', '4'], 1, "success: simulated"),
    ],
)
async def test_run_action(mode, cmd, retcode, expected, monkeypatch):
    job_db = mock.Mock()

    async def mock_run_proc_async(*args):
        return retcode, b'Submitted job: 42', b''

    monkeypatch.setattr(
        webservice.webservice, 'run_proc_async', mock_run_proc_async
    )
    ret = await run_action(job_db, cmd, mode, 1, 1, 1)
    assert ret.lower().startswith(expected)

@pytest.mark.asyncio
@pytest.mark.parametrize(
    'proposal_number, action, mock_proposal_status, expected_result',
    [
        (42, 'correct', 'A', 'upex-middle'),  # active
        ('42', 'dark', 'R', 'upex-high'),  # active
        (404, 'correct', 'FR', 'exfel'),  # finished and reviewed
        (404, 'dark', 'CLO', 'exfel'),  # closed
    ],
)
async def test_get_slurm_partition(proposal_number,
                                   action,
                                   mock_proposal_status,
                                   expected_result):

    response = mock.Mock()
    response.status_code = 200
    response.json = mock.Mock(return_value={'flg_beamtime_status': mock_proposal_status})
    client = mock.Mock()
    client.get_proposal_by_number_api = mock.Mock(
        return_value=response)

    ret = await get_slurm_partition(client, action, proposal_number)
    assert ret == expected_result


@pytest.mark.parametrize(
    'commissioning, run_age_days, expected_partition',
    [
        (False, 0, "upex-middle"),
        (True, 0, "upex-middle"),
        (False, config["correct"]["commissioning-max-age-days"], "upex-middle"),
        (True, config["correct"]["commissioning-max-age-days"], "upex-middle"),
        (False, config["correct"]["commissioning-max-age-days"]+1, "upex-middle"),
        (True, config["correct"]["commissioning-max-age-days"]+1, "exfel"),
    ]
)
@pytest.mark.asyncio
async def test_get_slurm_partition_run_age(
    commissioning: bool,
    run_age_days: int,
    expected_partition: str,
):
    # NOTE: non-zero at cycle index 4 (`str(cycle)[4]`) indicates commissioning
    cycle = 202231 if commissioning else 202201

    now = dt.datetime.now().astimezone()
    run_dt = now - dt.timedelta(days=run_age_days)
    run_str = run_dt.strftime("%Y-%m-%dT%H:%M:%S.%f%z")

    # Note that, as the fields are independent, this mocks both the
    # response for `get_proposal_by_number_api` AND `get_run_by_id_api`
    response = mock.Mock()
    response.status_code = 200
    response.json = lambda: {"flg_beamtime_status": "A", "begin_at": run_str}

    client = mock.Mock()
    client.get_proposal_by_number_api = mock.Mock(return_value=response)
    client.get_run_by_id_api = mock.Mock(return_value=response)

    ret = await get_slurm_partition(
        client, "correct", 1, run_id=1, cycle=cycle
    )

    assert ret == expected_partition


@pytest.mark.asyncio
@pytest.mark.parametrize(
    'cycle, num_jobs, expected_result',
    [
        ('202201', 0, 0), ('202201', 10, 3*10**2),  # user proposal
        ('202221', 0, 5), ('202221', 10, 5+3*10**2),  # commissioning
    ]
)
async def test_get_slurm_nice_values(fp, cycle, num_jobs, expected_result):
    """ Test get_slurm_nice values."""

    fp.register(
        ['squeue', '-h', '-o', '%.20j', '-p', 'upex-higher', '--me'],
        stdout='\n'.join(
            [f'correct_SPB_{i}' for i in range(num_jobs)] +
            [f'correct_FXE_{i}' for i in range(num_jobs*2)]).encode('ascii'),
        returncode=0)

    ret = await get_slurm_nice(
        'upex-higher', 'SPB', cycle, job_penalty=3, commissioning_penalty=5)
    assert ret == expected_result

@pytest.mark.asyncio
async def test_get_slurm_nice_fails(fp):
    """Test corner cases for get_slurm_nice."""

    # non-zero returncode
    fp.register(
        ['squeue', '-h', '-o', '%.20j', '-p', 'upex-higher', '--me'],
        stdout='', returncode=1)

    assert await get_slurm_nice('upex-higher', 'SPB', '202201') == 0

    # exfel is special
    fp.register(
        ['squeue', '-h', '-o', '%.20j', '-p', 'exfel', '--me'],
        stdout='\n'.join([f'correct_SPB_{i}' for i in range(10)]),
        returncode=0)

    assert await get_slurm_nice('exfel', 'SPB', '202201') == 0

